import sys
import os
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from multiprocessing import Pool

from eis1600.helper.repo import get_path_to_other_repo, read_files_from_readme, get_files_from_eis1600_dir, \
    write_to_readme
from eis1600.miu.methods import disassemble_text


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and os.path.isfile(input_arg):
            filepath, fileext = os.path.splitext(input_arg)
            if fileext != '.EIS1600':
                parser.error('You need to input a EIS1600 file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to disassemble EIS1600 file(s) into MIU file(s).
-----
Give a single EIS1600 file as input
or 
Use -e <EIS1600_repo> to batch process all EIS1600 files in the EIS1600 directory.
'''
    )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            '-e', '--eis1600_repo', type=str,
            help='Takes a path to the EIS1600 file repo and batch processes all files'
    )
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 file to process',
            action=CheckFileEndingAction
    )
    args = arg_parser.parse_args()

    verbose = args.verbose

    if args.input:
        infile = './' + args.input
        out_path = get_path_to_other_repo(infile, 'MIU')
        print(f'Disassemble {infile}')
        disassemble_text(infile, out_path, verbose)
        infiles = [infile.split('/')[-1]]
        path = out_path.split('data')[0]
        write_to_readme(path, infiles, '# Texts disassembled into MIU files\n')
    elif args.eis1600_repo:
        input_dir = args.eis1600_repo
        if not input_dir[-1] == '/':
            input_dir += '/'
        out_path = get_path_to_other_repo(input_dir, 'MIU')

        print(f'Disassemble EIS1600 files from the EIS1600 repo')
        files_list = read_files_from_readme(input_dir, '# Texts converted into `.EIS1600`\n')
        files_list_done = read_files_from_readme(input_dir, '# Texts disassembled into MIU files\n')
        files_list = [file for file in files_list if file not in files_list_done]
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600')
        if not infiles:
            print('There are no EIS1600 files to process')
            sys.exit()

        params = [(infile, out_path, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(disassemble_text, params).get()

        write_to_readme(out_path, infiles, '# Texts disassembled into MIU files\n')
    else:
        print(
                'Pass in a <uri.EIS1600> file to process a single file or use the -e option for batch processing'
        )
        sys.exit()

    print('Done')



import os
import re
import sys
import random
import subprocess


readme = """
===================================================================================
add detailed readme here to be printed out if YML file does not exist in the folder
...................................................................................
===================================================================================
"""

ymlTemplate = """

system: ... # options: mac, lin, win;
reviewer: eis100researcher # change this to your name;
path_to_kate: kate # add absolute path to Kate on your machine; or a working alias
files_to_open: 10 # the script will open 10 files by default; you can increase/decrease this value;
miu_main_path: ./data/ # the script should be run from inside the repository with MIU files
miu_specific_uri: ... # you can paste a URI of a specific book in order to focus on it

"""


def readCustomYML(text):
    text = text.strip()
    text = text.split("\n")

    dic = {}

    for t in text:
        t = t.split(":", maxsplit=1)
        key = t[0]
        val = re.sub("#.*$", "", t[1]).strip()

        dic[key] = val

    return dic


def generateDicOfMIUs(path):
    dic = {}
    for subdir, dirs, files in os.walk(path):
        for file in files:
            # process publication tf data
            if file.endswith(".EIS1600"):
                key = file
                value = os.path.join(subdir, file)
                dic[key] = value
    return dic


def sampleMIUs(dictionary, filterParameter, filesToOpen):
    selection = []

    for k in dictionary.keys():
        if filterParameter in k:
            selection.append(k)

    finalSelection = random.sample(selection, filesToOpen*3)

    return finalSelection


def main():
    if os.path.exists("file_picker.yml"):
        with open("file_picker.yml", "r", encoding="utf8") as f1:
            ymlText = f1.read().strip()
            ymlDic = readCustomYML(ymlText)

            print("===================================================================================")
            for k, v in ymlDic.items():
                print(k, ": ", v)
            print("===================================================================================")
    else:
        with open("file_picker.yml", "w", encoding="utf8") as f1:
            f1.write(ymlTemplate)

        print("===================================================================================")
        print("YML FILE HAS BEEN GENERATED; PLEASE UPDATE THE YML FILE AS DESCRIBED ==============")
        print("===================================================================================")
        print(ymlTemplate)
        print("===================================================================================")
        print("YML FILE MUST BE COMPLETED FOR THE FILE PICKER TO WORK PROPERLY. COMPLETE AND RERUN")
        print("===================================================================================")

        # print(readme)

        sys.exit()

    ################################################################################################
    # FINALIZING FILTERING PARAMETERS ##############################################################
    ################################################################################################

    if re.search(r"^\W+$", ymlDic["miu_specific_uri"]):
        filterParameter = "."
    else:
        filterParameter = ymlDic["miu_specific_uri"]

    filesToOpen = int(ymlDic["files_to_open"])
    pathToKate = ymlDic["path_to_kate"]
    path = ymlDic["miu_main_path"]
    reviewer = ymlDic["reviewer"]
    osVar = ymlDic["system"]

    ################################################################################################
    # LOADING MIU DATA, RANDOMLY SELECTING, OPENING FILES WITH KATE ################################
    ################################################################################################

    dictionary = generateDicOfMIUs(path)
    with open("selected_files.txt", "r", encoding="utf8") as f1:
        mius_to_open = f1.read().split("\n")

    ################################################################################################
    # OPEN MIUS from THE SELECTED SAMPLE ###########################################################
    ################################################################################################

    for miu in mius_to_open:
        with open(dictionary[miu], "r+", encoding="utf8") as ft:
            MIUcontent = ft.read()

            if re.search(r"reviewed\s+:\s+REVIEWED", MIUcontent):
                # insert the name of the reviewer
                if re.search(r"reviewer\s+:", MIUcontent):
                    MIUcontent = re.sub(r"(reviewed\s+:)[^\n]+\n", r"\12\n", MIUcontent)
                    MIUcontent = re.sub(r"(reviewer\s+:[^\n]+)\n", r"\1, %s\n" % reviewer, MIUcontent)

                # inserting the name of the reviewer; the reviewer will have to only change the status tag
                # upon saving and closing
            ft.seek(0)
            ft.write(MIUcontent)
            ft.truncate()

        # NOW OPENING IN KATE FOR REVIEW
        if osVar == "mac":
            # open on mac
            # lineToRun = "open -a %s %s" % (pathToKate, dictionary[miu])
            # os.system(lineToRun)
            subprocess.run(['open', '-a', pathToKate, dictionary[miu]])
        elif osVar == "lin":
            subprocess.run([pathToKate, dictionary[miu]])
        elif osVar == "win":
            subprocess.run([pathToKate, dictionary[miu]])
        else:
            print("Operating system is incorrect. Use: mac, lin, or win")
            sys.exit()

        print(','.join([miu, reviewer, 'in progress']))


    print("======================================================================================")
    print("REVIEW MIU FILES IN KATE; WHEN DONE, CHANGE STATUS TO `REVIEWED`, SAVE, AND CLOSE ====")
    print("======================================================================================")



import re

AR_LETTERS_CHARSET = frozenset(
        u'\u0621\u0622\u0623\u0624\u0625\u0626\u0627'
        u'\u0628\u0629\u062a\u062b\u062c\u062d\u062e'
        u'\u062f\u0630\u0631\u0632\u0633\u0634\u0635'
        u'\u0636\u0637\u0638\u0639\u063a\u0640\u0641'
        u'\u0642\u0643\u0644\u0645\u0646\u0647\u0648'
        u'\u0649\u064a\u0671\u067e\u0686\u06a4\u06af'
)
AR_STR = r'[' + u''.join(AR_LETTERS_CHARSET) + ']+'
AR_STR_AND_TAGS = r'[' + u''.join(AR_LETTERS_CHARSET) + 'a-zA-ZÜ0-9]+'
WORD = r'(?:\s' + AR_STR + ')'

# EIS1600 mARkdown
UID = r'_ء_(#)?=(?P<UID>\d{12})= '
UID_PATTERN = re.compile(UID)
MIU_UID = r'_ء_#=(?P<UID>\d{12})= '
MIU_UID_PATTERN = re.compile(MIU_UID)
HEADER_END_PATTERN = re.compile(r'(#META#Header#End#)\n')
MIU_HEADER = r'#MIU#Header#'
MIU_HEADER_PATTERN = re.compile(MIU_HEADER)
HEADING_PATTERN = re.compile(UID + r'(?P<level>[|]+) (?P<heading>.*)\n')
EMPTY_PARAGRAPH_PATTERN = re.compile(UID + r'::UNDEFINED:: ~')
EMPTY_FIRST_PARAGRAPH_PATTERN = re.compile(r'_ء_#=\d{12}=')
PAGE_TAG = r' ?(?P<page_tag>PageV\d{2}P\d{3,})'
PAGE_TAG_PATTERN = re.compile(PAGE_TAG)
ONLY_PAGE_TAG = UID + r'::UNDEFINED:: ~\n' + PAGE_TAG
ONLY_PAGE_TAG_PATTERN = re.compile(ONLY_PAGE_TAG)
PAGE_TAG_IN_BETWEEN_PATTERN = re.compile(
        AR_STR + r' ?' + r'\n\n' + ONLY_PAGE_TAG + r'\n\n' + r'_ء_=\d{12}= ::[A-Z]+:: ~\n' + AR_STR
)

MIU_TAG_PATTERN = re.compile(r'(' + MIU_UID + r'[^\n]+)')
SECTION_TAG = r'_ء_=\d{12}= ::[A-Z]+:: ~'
SECTION_PATTERN = re.compile(SECTION_TAG)
SECTION_SPLITTER_PATTERN = re.compile(r'\n\n(' + SECTION_TAG + ')\n(?:_ء_)?')
TAG_PATTERN = re.compile(r'Ü?(?:[a-zA-Z0-9_%~]+(?:\.[a-zA-Z0-9_%~]+)?)|' + PAGE_TAG)
NOR_DIGIT_NOR_AR_STR = r'[^\d\n' + u''.join(AR_LETTERS_CHARSET) + ']+?'
TAG_AND_TEXT_SAME_LINE_PATTERN = re.compile(
        r'(_ء_#=\d{12}= [$@]+(?:' + NOR_DIGIT_NOR_AR_STR + r')?(?:\d+)?(?:' + NOR_DIGIT_NOR_AR_STR + r')?) ('
                                                                                                               r'(?:\( ?)?' +
        AR_STR + r')'
)
MIU_TAG_AND_TEXT_PATTERN = re.compile(r'(' + MIU_UID + r'[$@]+(?: RAW)?(?: \d+)?)\n((?:\( ?)?' + AR_STR + r')')

# EIS1600 light mARkdown
HEADING_OR_BIO_PATTERN = re.compile(r'# [|$]+')
MIU_LIGHT_OR_EIS1600_PATTERN = re.compile(r'#|_ء_#')

# Fix mARkdown files
SPACES_CROWD_PATTERN = re.compile(r' +')
NEWLINES_CROWD_PATTERN = re.compile(r'\n{3,}')
SPACES_AFTER_NEWLINES_PATTERN = re.compile(r'\n +')
POETRY_PATTERN = re.compile(
        r'# (' + AR_STR_AND_TAGS + '(?: ' + AR_STR_AND_TAGS + ')* %~% ' + AR_STR_AND_TAGS + '(?: ' +
        AR_STR_AND_TAGS +
        r')*) ?'
)
BELONGS_TO_PREV_PARAGRAPH_PATTERN = re.compile(r'\n(.{1,10})\n')
PAGE_TAG_ON_NEWLINE_PATTERN = re.compile(r'\n' + PAGE_TAG)
PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN = re.compile(
        '(' + AR_STR + ' ?)' + r'\n\n' + PAGE_TAG + r'\n\n' + '(' + AR_STR +
        ')'
)
BIO_CHR_TO_NEWLINE_PATTERN = re.compile(
        r'(# [$@]+(?:' + NOR_DIGIT_NOR_AR_STR + r')?(?:\d+)?(?:' + NOR_DIGIT_NOR_AR_STR + r')?) ((?:\( ?)?'
        + AR_STR + r')'
)

# Fixed poetry old file path pattern
FIXED_POETRY_OLD_PATH_PATTERN = re.compile(r'/Users/romanov/_OpenITI/_main_corpus/\w+/data/')



"""

This module contains functions to work with the EIS1600 text repo. This includes retrieving files from the repo,
as well as reading and writing to the README file of that repo.

Functions:
:function write_to_readme(path, files, which, ext=None, checked=False, remove_duplicates=False):
:function read_files_from_readme(path, which):
:function update_texts_fixed_poetry_readme(path, which):
:function get_files_from_eis1600_dir(path, file_list, file_ext_from, file_ext_to):
:function travers_eis1600_dir(path, file_ext_from, file_ext_to): Discontinued
"""

from glob import glob
from os.path import split, splitext
from typing import List, Literal, Optional

from eis1600.markdown.re_patterns import FIXED_POETRY_OLD_PATH_PATTERN


def get_entry(file_name: str, checked_entry: bool) -> str:
    """Formats README entry for that file_name.

    Only used internally.
    :param str file_name: The name of the file whose entry is added to the README
    :param bool checked_entry: Bool indicating if the checkbox of that entry is ticked or not
    :return str: The formatted entry which can be added to the README file
    """

    x = 'x' if checked_entry else ' '
    return '- [' + x + '] ' + file_name


def write_to_readme(path: str, files: List[str], which: str, ext: Optional[str] = None, checked: bool = False) -> None:
    """Write list of successfully processed files to the README.

    Write processed files to the respective section in the README, sorted into existing lists.

    :param str path: The root of the text repo, path to the README
    :param list[str] files: List of files to write to the respective section in the README
    :param str which: The section heading from the README indicating the section to write the list of files to.
    :param str or None ext: File extension of the files at the end of the process, optional.
    :param bool checked: Indicator if the checkboxes of the files are ticked, defaults to False.
    """

    file_list = []
    try:
        with open(path + 'README.md', 'r', encoding='utf8') as readme_h:
            out_file_start = ''
            out_file_end = ''
            checked_boxes = False
            line = next(readme_h)
            # Find section in the README to write to by finding the corresponding header line
            while line != which:
                out_file_start += line
                line = next(readme_h)
            out_file_start += line
            out_file_start += next(readme_h)
            line = next(readme_h)
            # Read existing entries from that section
            while line and line != '\n':
                if line.startswith('- ['):
                    checked_boxes = True
                    md, file = line.split('] ')
                    file_list.append((file, md == '- [x'))
                    line = next(readme_h, None)
                else:
                    file_list.append(line[2:])
                    line = next(readme_h, None)
            while line:
                out_file_end += line
                line = next(readme_h, None)

        # Change the file ending for files which have been processed if necessary (if new file ending is given)
        for file in files:
            file_path, uri = split(file)
            if ext:
                uri, _ = splitext(uri)
            else:
                uri, ext = splitext(uri)
            if checked_boxes:
                file_list.append((uri + ext + '\n', checked))
            else:
                file_list.append(uri + ext + '\n')

        # Remove duplicates
        file_list = list(set(file_list))
        # Sort list of all entries (old and new)
        file_list.sort()

        # Write new list to section in the readme
        with open(path + 'README.md', 'w', encoding='utf8') as readme_h:
            readme_h.write(out_file_start)
            if checked_boxes:
                readme_h.writelines([get_entry(file, checked_entry) for file, checked_entry in file_list])
            else:
                readme_h.writelines(['- ' + file for file in file_list])
            readme_h.write(out_file_end)

    except StopIteration:
        # Fallback option - if anything goes wrong at least print the list of changed files to a log
        file_list = []
        for file in files:
            file_path, uri = split(file)
            uri, ext = splitext(uri)
            file_list.append(uri + '.EIS1600\n')
        with open(path + 'FILE_LIST.log', 'w', encoding='utf8') as file_list_h:
            file_list_h.writelines(file_list)

        print(f'Could not write to the README file, check {path + "FILE_LIST.log"} for changed files')


def read_files_from_readme(path: str, which: str, only_checked: Optional[bool] = True) -> List[str]:
    """Get the list of files from the README to process further.

    Get the list of files from the README which are to be processed in further steps.
    :param str path: The root of the text repo, path to the README
    :param str which: The section heading from the README indicating the section from which to read the file list from.
    :param bool only_checked: If True, only read those lines with a ticked checkbox, defaults to True.
    :return list[str]: List of URIs from files to process further
    """

    file_list = []
    try:
        with open(path + 'README.md', 'r', encoding='utf8') as readme_h:
            line = next(readme_h)
            # Find section in the README to read from
            while line != which:
                line = next(readme_h)
            next(readme_h)
            line = next(readme_h)
            # Read files from that section
            while line and line != '\n':
                if line.startswith('- ['):
                    if only_checked:
                        # Only read those files which have been checked
                        if line.startswith('- [x'):
                            md, file = line.split('] ')
                            file_list.append(file[:-1])
                    else:
                        md, file = line.split('] ')
                        file_list.append(file[:-1])
                else:
                    file_list.append(line[2:-1])

                line = next(readme_h, None)
    except StopIteration:
        print(f'The README.md file does not seem to contain a "{which[:-1]}" section')

    return file_list


def update_texts_fixed_poetry_readme(path: str, which: str) -> None:
    """Update list of texts with fixed poetry in the README.

    Read list of texts with fixed poetry from the text file in the scipt folder and update the respective list in the
    README.

    :param str path: Path to the text directory root.
    :param str which: The section heading from the README indicating the section where the texts with fixed poetry
    are listed.
    """

    # Read the list of files with fixed poetry from other file and write it to the README
    with open(path + 'scripts/poetry_fixed.txt', 'r', encoding='utf8') as readme_h:
        files_text = readme_h.read()
    files_text = FIXED_POETRY_OLD_PATH_PATTERN.sub('', files_text)
    file_list = files_text.split('\n')
    write_to_readme(path, file_list, which)


def get_files_from_eis1600_dir(
        path: str, file_list: List[str], file_ext_from: List[str] or str, file_ext_to: Optional[str] = None
) -> List[str]:
    """Get list of files to process from the EIS1600 text repo.

    Get list of the files with exact path from list of URIs. Do not select those files which have already been
    processed (those files already exist with the new file extension).

    :param str path: Path to the text directory root.
    :param list[str] file_list: List of URIs of files.
    :param str or list[str] file_ext_from: File extension(s) the unprocessed files have.
    :param str file_ext_to: File extension already processed files have, optional.
    :return list[str]: List of all files to process with exact path, not containing those files which have already
    been processed.
    """

    path += 'data/'
    files = []
    for file in file_list:
        author, work, text = file.split('.')[:3]
        file_path = path + '/'.join([author, '.'.join([author, work]), '.'.join([author, work, text])]) + '.'
        if file_ext_to and not glob(file_path + file_ext_to):
            # Only do if the target file does not exist
            if type(file_ext_from) == list:
                for ext in file_ext_from:
                    tmp = glob(file_path + ext)
                    if tmp:
                        files.extend(tmp)
            else:
                files.extend(glob(file_path + file_ext_from))
        elif not file_ext_to:
            # Do if file ending stays the same
            files.extend(glob(file_path + file_ext_from))
    return files


def get_path_to_other_repo(infile: str, which: Literal['MIU', 'TEXT']) -> str:
    """

    :param str infile: file which gives URI of the text.
    :param Literal which: Indicating which repo you want to get the path to, accepts 'MIU' or 'TEXT'.
    :return str: path to the same URI in the requested repo
    """
    out_path = '../'

    if 'data' in infile:
        out_path += infile.split('data')[0][2:]
    elif infile == './':
        pass
    else:
        depth = len(infile.split('/'))
        print(depth)
        if depth == 1:
            out_path += '../../../'
        elif depth == 2:
            out_path += '../../'
        elif depth == 3:
            out_path += '../'
        else:
            out_path += '../'

    if which == 'MIU':
        out_path += 'OpenITI_EIS1600_MIUs/data/'
    else:
        out_path += 'OpenITI_EIS1600_Texts/data/'

    return out_path



from glob import glob
from os.path import splitext, split, exists
from typing import List, Optional
from pathlib import Path

from eis1600.dates.methods import date_annotate_miu_text
from eis1600.miu.HeadingTracker import HeadingTracker
from eis1600.preprocessing.methods import get_yml_and_MIU_df, write_updated_miu_to_file
from eis1600.nlp.utils import camel2md_as_list, annotate_miu_text
from eis1600.miu.yml_handling import create_yml_header, extract_yml_header_and_text
from eis1600.markdown.re_patterns import HEADER_END_PATTERN, HEADING_PATTERN, MIU_UID_PATTERN, PAGE_TAG_PATTERN, \
    UID_PATTERN


def disassemble_text(infile: str, out_path: str, verbose: Optional[bool] = None) -> None:
    """Disassemble text into MIU files.

    Retrieve MIU files by disassembling the text based on the EIS1600 mARkdown.
    :param str infile: Path to the file which is to be disassembled.
    :param str out_path: Path to the MIU repo.
    :param bool verbose: If True outputs a notification of the file which is currently processed, optional.
    """

    heading_tracker = HeadingTracker()
    path, uri = split(infile)
    uri, ext = splitext(uri)
    author, work, text = uri.split('.')
    path = out_path + '/'.join([author, '.'.join([author, work])]) + '/'
    # TODO fix ids file path
    ids_file = out_path + uri + '.IDs'
    yml_file = out_path + uri + '.STATUS.yml'
    miu_dir = Path(path + 'MIUs/')
    uid = ''
    miu_text = ''

    if verbose:
        print(f'Disassemble {uri}')

    miu_dir.mkdir(parents=True, exist_ok=True)
    miu_uri = miu_dir.__str__() + '/' + uri + '.'

    with open(infile, 'r', encoding='utf8') as text:
        with open(ids_file, 'w', encoding='utf8') as ids_tree:
            for text_line in iter(text):
                if HEADER_END_PATTERN.match(text_line):
                    uid = 'header'
                    miu_text += text_line
                    with open(miu_uri + uid + '.EIS1600', 'w', encoding='utf8') as miu_file:
                        miu_file.write(miu_text + '\n')
                        ids_tree.write(uid + '\n')
                    miu_text = ''
                    uid = 'preface'
                    next(text)  # Skip empty line after header
                elif MIU_UID_PATTERN.match(text_line):
                    if HEADING_PATTERN.match(text_line):
                        m = HEADING_PATTERN.match(text_line)
                        heading_text = m.group('heading')
                        if PAGE_TAG_PATTERN.search(heading_text):
                            heading_text = PAGE_TAG_PATTERN.sub('', heading_text)
                        heading_tracker.track(len(m.group('level')), heading_text)
                    if miu_text:
                        # Do not create a preface MIU file if there is no preface
                        with open(miu_uri + uid + '.EIS1600', 'w', encoding='utf8') as miu_file:
                            miu_file.write(miu_text + '\n')
                            ids_tree.write(uid + '\n')
                    uid = UID_PATTERN.match(text_line).group('UID')
                    miu_text = create_yml_header(heading_tracker.get_curr_state())
                    miu_text += text_line
                else:
                    miu_text += text_line
            # last MIU needs to be written to file when the for-loop is finished
            with open(miu_uri + uid + '.EIS1600', 'w', encoding='utf8') as miu_file:
                miu_file.write(miu_text + '\n')
                ids_tree.write(uid + '\n')

    with open(yml_file, 'w', encoding='utf8') as status_file:
        status_file.write('STATUS   : DISASSEMBLED')


def reassemble_text(infile: str, out_path: str, verbose: Optional[bool] = None) -> None:
    """Reassemble text from MIU files.

    Reassemble text from MIU files.
    :param str infile: Path to the IDs file of the text to reassemble from MIU files.
    :param str out_path: Path to the TEXT repo.
    :param bool verbose: If True outputs a notification of the file which is currently processed, optional.
    """
    file_path, uri = split(infile)
    file_path += '/'
    uri, ext = splitext(uri)
    author, work, text = uri.split('.')
    path = out_path + '/'.join([author, '.'.join([author, work])]) + '/' + uri
    ids = []

    if verbose:
        print(f'Reassemble {uri}')

    with open(file_path + uri + '.IDs', 'r', encoding='utf-8') as ids_file:
        ids.extend([line[:-1] for line in ids_file.readlines()])

    with open(path + '.EIS1600', 'w', encoding='utf-8') as text_file:
        with open(file_path + uri + '.YAMLDATA.yml', 'w', encoding='utf-8') as yml_data:
            for i, miu_id in enumerate(ids):
                miu_file_path = file_path + 'MIUs/' + uri + '.' + miu_id + '.EIS1600'
                with open(miu_file_path, 'r', encoding='utf-8') as miu_file_object:
                    yml_header, text = extract_yml_header_and_text(miu_file_object, i == 0)
                text_file.write(text)
                yml_data.write('#' + miu_id + '\n---\n' + yml_header + '\n\n')


def get_mius(infile: str) -> List[str]:
    """Get a list of paths to all MIU files from the infile text (including the HEADER MIU)

    :param infile: URI of text for which all MIU files are retrieved
    :return List[str]: A List of path to all the MIU files from the text
    """
    file_path, uri = split(infile)
    uri, ext = splitext(uri)
    file_path += '/'
    ids = []
    mius = []

    with open(infile, 'r', encoding='utf-8') as ids_file:
        ids.extend([line[:-1] for line in ids_file.readlines()])

    for i, miu_id in enumerate(ids):
        mius.extend(glob(file_path + 'MIUs/' + uri + '.' + miu_id + '.EIS1600'))

    return mius


def annotate_miu_file(path: str, tsv_path=None, output_path=None, force_annotation=False):
    if output_path is None:
        output_path = path
    if tsv_path is None:
        tsv_path = path.replace('.EIS1600', '.tsv')

    # if the file is already annotated, do nothing
    if exists(tsv_path) and not force_annotation:
        return

    with open(path, 'r+', encoding='utf-8') as miu_file_object:
        # 1. open miu file and disassemble the file to its parts
        yml_header, df = get_yml_and_MIU_df(miu_file_object)

        # 2. annotate NEs and lemmatize
        df['NER_LABELS'], df['LEMMAS'], df['POS_TAGS'] = annotate_miu_text(df)

        # 3. convert cameltools labels format to markdown format
        df['NER_TAGS'] = camel2md_as_list(df['NER_LABELS'].tolist())

        # 4. annotate dates
        df['DATE_TAGS'], yml_header = date_annotate_miu_text(df[['TOKENS']], yml_header)

        # 5. save csv file
        df.to_csv(tsv_path, index=False, sep='\t')

        # 6. reconstruct the text and save it to the output file
        if output_path == path:
            write_updated_miu_to_file(
                miu_file_object, yml_header, df[['SECTIONS', 'TOKENS', 'TAGS_LISTS', 'NER_TAGS',
                                                 'DATE_TAGS']]
                )
        else:
            with open(output_path, 'w', encoding='utf-8') as out_file_object:
                write_updated_miu_to_file(
                        out_file_object, yml_header, df[['SECTIONS', 'TOKENS', 'TAGS_LISTS', 'NER_TAGS',
                                                         'DATE_TAGS']]
                )



from copy import copy

from typing import Optional, List, Set

from random import randint


class UIDs:
    """Generates a set of UIDs which can be used by all instances of UIDs.

    Generates one large set of UIDs when this class is imported. Each instance reuses that set, but subtracts already
    used UIDs. Saves time as the set only needs to be generated once and is reused and adapted in instances.

    :param List[int] existing_uids: List of UIDs already used in the current text, optional.
    :ivar Set[int] ids: Set of UIDs which does not contain already used UIDs.
    """

    @staticmethod
    def _generate_ids() -> Set[int]:
        """Generates a set of 5000000 UIDs.

        :return Set[int]: A set of 5000000 UIDs.
        """

        ids = []
        for i in range(0, 5000000):
            ids.append(randint(400000000000, 999999999999))
        return set(ids)

    IDS = _generate_ids.__func__()
    """Set of 5000000 random UIDs generated once and shared by all instances of this class."""

    def __init__(self, existing_uids: Optional[List[int]] = None) -> None:
        """Constructor method, creates an individual set uids per instance.

        Adaptes the pre-generated set of UIDs to the current instance by subtracting already existing UIDs,
        if this list was given to the constructor. Otherwise it uses a copy of the static IDS set.
        :param List[int] existing_uids: List of already used UIDs in the current text.
        """

        if existing_uids:
            self.ids = UIDs.IDS.difference(existing_uids)
        else:
            self.ids = copy(UIDs.IDS)

        self.iter = self.ids.__iter__()

    def get_uid(self) -> int:
        """Returns an individual UID."""

        return next(self.iter)




from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='eis1600',
      version='0.7.1',
      description='EIS1600 project tools and utilities',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/EIS1600/eis1600-pkg',
      author='Lisa Mischer',
      author_email='mischer.lisa@gmail.com',
      license='MIT License',
      packages=['eis1600',
                'eis1600.dates',
                'eis1600.helper',
                'eis1600.markdown',
                'eis1600.miu',
                'eis1600.nlp'],
      entry_points={
          'console_scripts': [
                  'convert_mARkdown_to_EIS1600TMP = eis1600.markdown.convert_mARkdown_to_EIS1600TMP:main',
                  'disassemble_into_miu_files = eis1600.miu.disassemble_into_miu_files:main',
                  'insert_uids = eis1600.markdown.insert_uids:main',
                  'miu_random_revisions = eis1600.helper.miu_random_revisions:main',
                  'ner_annotate_mius = eis1600.nlp.ner_annotate_mius:main [NER]',
                  'reassemble_from_miu_files = eis1600.miu.reassemble_from_miu_files:main',
                  'update_uids = eis1600.markdown.update_uids:main',
                  'xx_update_uids_old_process = eis1600.markdown.update_uids_old_process:main'
          ],
      },
      python_requires='>=3.7, <3.9',
      install_requires=['openiti', 'pandas', 'tqdm', 'p_tqdm'],
      extras_require={'NER': ['camel-tools']},
      classifiers=['Programming Language :: Python :: 3',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Development Status :: 1 - Planning',
                   'Intended Audience :: Science/Research']
      )
